

#include "knowledgebase.h"
#include "parser/pl_lexer.h"
#include "parser/pl_parser.hpp"

using namespace std;

namespace Prolog
{

    Term::~Term()
    {
    }

    Term* Term::Clone(int& VariableIndex)
    {
        map<string, string> IndexedVariables;
        this->Clone(VariableIndex, IndexedVariables);
    }

    void Term::ReplaceVariables(Term** t, string* Name, Term* repl)
    {
        CompositeTerm* ct = dynamic_cast<CompositeTerm*>(*t);
        if(ct != NULL)
        {
            if(ct->Children != NULL)
                for(Terms::iterator i = ct->Children->begin(); i != ct->Children->end(); i++)
                    ReplaceVariables(&(*i), Name, repl);
        }
        else
        {
            Variable* v = dynamic_cast<Variable*>(*t);
            if(v != NULL && *v->Name == *Name)
            {
                Term* temp = repl->Clone();
                delete *t;
                *t = temp;
            }
        }
    }

    // =============================================================================

    CompositeTerm::CompositeTerm(string* Head, Terms* Children)
    {
        this->Head = Head;
        this->Children = Children;
    }

    CompositeTerm::~CompositeTerm()
    {
        delete Head;
        if(Children != NULL)
        {
            for(Terms::iterator i = Children->begin(); i != Children->end(); i++)
                delete *i;
            delete Children;
        }
    }

    void CompositeTerm::Portray(ostream& os)
    {
        os << *Head;
        if(Children != NULL)
        {
            os << '(';
            for(Terms::iterator i = Children->begin(); i != Children->end(); i++)
            {
                (*i)->Portray(os);
                if(i+1 != Children->end())
                    os << ',';
            }
            os << ')';
        }
    }

    bool CompositeTerm::ContainsVariable(string* Name)
    {
        if(Children != NULL)
            for(Terms::iterator i = Children->begin(); i != Children->end(); i++)
                if((*i)->ContainsVariable(Name))
                    return true;
        return false;
    }

    Term* CompositeTerm::Clone()
    {
        Terms* children = Children == NULL ? NULL : new Terms(Children->size());
        if(Children != NULL)
            for(Terms::iterator i = Children->begin(); i != Children->end(); i++)
                children->push_back((*i)->Clone());

        return new CompositeTerm(new string(*this->Head), children);
    }

    Term* CompositeTerm::Clone(int& VariableIndex, map<string,string>& IndexedVariables)
    {
        Terms* children = Children == NULL ? NULL : new Terms;
        if(Children != NULL)
            for(Terms::iterator i = Children->begin(); i != Children->end(); i++)
                children->push_back((*i)->Clone(VariableIndex, IndexedVariables));

        return new CompositeTerm(new string(*this->Head), children);
    }

    // =============================================================================

    Variable::Variable(string* Name)
    {
        this->Name = Name;
    }

    Variable::~Variable()
    {
        delete Name;
    }

    void Variable::Portray(ostream& os)
    {
        os << *Name;
    }

    bool Variable::ContainsVariable(string* Name)
    {
        return (*this->Name) == *Name;
    }

    Term* Variable::Clone()
    {
        return new Variable(new string(*Name));
    }


    Term* Variable::Clone(int& VariableIndex, map<string,string>& IndexedVariables)
    {
        // new variable -> generate a new replacement name (e.g. "_G1234")
        if(IndexedVariables.find(*Name) == IndexedVariables.end())
        {
            stringstream s;
            s << "_G" << VariableIndex;
            VariableIndex++;
            IndexedVariables[*Name] = s.str();
        }

        return new Variable(new string(IndexedVariables[*Name]));
    }




    // =========================================================================

    Clause::Clause(Term* Head, Terms* Block)
    {
        this->Head = Head;
        this->Block = Block;
    }

    Clause::~Clause()
    {
        delete Head;
        for(Terms::iterator i = Block->begin(); i != Block->end(); i++)
            delete *i;
        delete Block;
    }

    void Clause::Portray(ostream& os)
    {
        Head->Portray(os);
        if(Block->size() > 0)
            os << " :-\n";
        else
            os << ".\n";


        for(Terms::iterator i = Block->begin(); i != Block->end(); i++)
        {
            os << "  ";
            (*i)->Portray(os);
            if(i+1 != Block->end())
                os << ",\n";
            else
                os << ".\n";
        }

    }

    Clause* Clause::Clone(int& VariableIndex)
    {
        map<string, string> IndexedVariables;

        Term* HeadClone = Head->Clone(VariableIndex, IndexedVariables);

        Terms* BlockClone = new Terms;
        for(Terms::iterator i = Block->begin(); i != Block->end(); i++)
            BlockClone->push_back((*i)->Clone(VariableIndex, IndexedVariables));

        return new Clause(HeadClone, BlockClone);
    }

    // =========================================================================

    KnowledgeBase::KnowledgeBase()
    {
        goals = NULL;
        clauses = new Clauses;
    }

    KnowledgeBase::~KnowledgeBase()
    {
        for(Clauses::iterator i = clauses->begin(); i != clauses->end(); i++)
            delete *i;
        delete clauses;

        if(goals != NULL)
        {
            for(Goals::iterator i = goals->begin(); i != goals->end(); i++)
            {
                for(Goal::iterator j = (*i)->begin(); j != (*i)->end(); j++)
                    delete *j;
                delete *i;
            }
            delete goals;
        }
    }

    void KnowledgeBase::AddClause(Clause* clause)
    {
        this->clauses->push_back(clause);
    }

    void KnowledgeBase::Portray(ostream& os)
    {
        for(Clauses::iterator i = clauses->begin(); i != clauses->end(); i++)
            (*i)->Portray(os);
    }

    KnowledgeBase* KnowledgeBase::Parse(istream& is)
    {
        PLLexer lexer(is);
        KnowledgeBase* result = NULL;
        PLParser parser(lexer, &result);
        parser.parse();
        return result;
    }

}